from datetime import datetime
from abc import ABC, abstractmethod
from typing import Tuple, Optional


class Strategy(ABC):
    """Parses a string into a datetime and assigns it a confidence level.

    TODO: This design is completely arbitrary and the first thing that came
          to my mind.
    Confidence level guidance:
    (-infty, 0]
        Bad parse (consider returning None)
    (0, 25]
        Unlikely fuzzy match; some fixing/coercion has been applied
    (25, 50]
        Likely fuzzy match; a common fix/coercion has been applied (e.g. a
        common typo has been fixed)
    (50, 100]
        Exact match; the input exactly fits the pattern
    (100, infty)
        Special cases + headroom for hacks and overrides


    TODO: Allow users to prefer/pick/override strategies via labels
    TODO: Think about better way to prevent weird interactions and endless
          global tweaking
    TODO: Maybe we should be able to return multiple parses from a single
          strategy?
    TODO: Maybe strategies should have a finite number of confidences so that
          we can enumerate all the comparisons?
    TODO: Probably strategies themselves should be weighted based on the
          frequency of use?
    """
    @abstractmethod
    def parse(self, timespec: str) -> Optional[Tuple[float, datetime]]:
        pass


class TimestampStrategy(Strategy):
    @abstractmethod
    def parse_timestamp(self, timestamp: float):
        pass

    def parse(self, timespec):
        try:
            # TODO: Handle non-decimals etc.
            return self.parse_timestamp(float(timespec))
        except ValueError:
            return None


class SecondsTimestampStrategy(TimestampStrategy):
    def parse_timestamp(self, timestamp: float):
        # Most second timestamps are integers
        return [
            100. if int(timestamp) == timestamp else 70.,
            datetime.fromtimestamp(timestamp)
        ]


class MillisTimestampStrategy(TimestampStrategy):
    def parse_timestamp(self, timestamp: float):
        # Prefer integers, prefer multiples of 1000
        if int(timestamp) == timestamp:
            confidence = 100. if timestamp % 1e3 == 0 else 90.
        else:
            confidence = 70.

        return [confidence, datetime.fromtimestamp(timestamp / 1e3)]


class MicrosTimestampStrategy(TimestampStrategy):
    def parse_timestamp(self, timestamp: float):
        # Prefer integers, prefer multiples of 1000000
        if int(timestamp) == timestamp:
            confidence = 100. if timestamp % 1e6 == 0 else 90.
        else:
            confidence = 70.

        return [confidence, datetime.fromtimestamp(timestamp / 1e6)]



from datetime import timedelta
import math

from wttime.strategy import *


class Parser:
    strategies = [
        SecondsTimestampStrategy(),
        MillisTimestampStrategy(),
        MicrosTimestampStrategy()
    ]

    @staticmethod
    def instant_likelihood(now: datetime, instant: datetime) -> float:
        """Given a time, estimates the likelihood (0, 1] that this is a real
        timestamp.

        The following assumptions are used:
        - Most timestamps are in the past by the point they are inspected
        - Most timestamps of interest are in the recent past
        - Timestamps before epoch are unlikely in computer systems
        - Dates in the future are likely relatively close (TTLs, durations)

        This is implemented as a combination of linear and logistic functions:
        - (-infty, epoch start] - half-sigmoid -> (0, 0.2]
        - (epoch start, T] - linear -> (0.2, 1]
        - (T, T + 60 days] - linear -> (1, 0.3]
        - (T + 60 days, infty) - half-sigmoid -> (0.3, 0)
        """
        def logcurve(max_value, midpoint, slope, x):
            return max_value / (1 + math.exp(-slope * (x - midpoint)))

        y2k = datetime(2000, 1, 1).timestamp()
        horizon = timedelta(days=60)
        instant_secs = instant.timestamp()

        if instant <= datetime.fromtimestamp(0):
            return logcurve(0.2 * 2, 0, 1. / y2k, instant_secs)
        elif instant <= now:
            return 0.2 + instant_secs / now.timestamp() * 0.8
        elif instant <= now + horizon:
            return 1 - 0.7 * (instant_secs - now.timestamp()) / \
                   horizon.total_seconds()
        else:
            return 0.3 - logcurve(0.3, (now + horizon).timestamp(), 1. / y2k,
                                  instant_secs)

    @staticmethod
    def with_likelihood(now: datetime, parse_result):
        if parse_result:
            confidence, result = parse_result
            return confidence * Parser.instant_likelihood(now, result), result

    def parse(self, now, timespec: str) -> Tuple[float, datetime]:
        parses = [
            Parser.with_likelihood(now, strategy.parse(timespec))
            for strategy in self.strategies
        ]
        guesses = list(filter(None, parses))
        guesses.sort(key=lambda g: g[0], reverse=True)
        if guesses:
            return guesses[0]



from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

setup(name='wttime',
      version='0.1.1',
      description='Smart timestamp utility',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/PJK/wttime/actions',
      author='Pavel Kalvoda',
      author_email='me@pavelkalvoda.com',
      license='MIT',
      packages=['wttime'],
      scripts=['bin/wttime'],
      install_requires=[
            'click',
            'python-dateutil'
      ],
      tests_require=['pytest'],
      zip_safe=False)
